
-------------------------------------------------------------------------------
-- Optimization
-------------------------------------------------------------------------------

local C_Club = C_Club
local coroutine = coroutine
local ipairs = ipairs

-------------------------------------------------------------------------------
-- Namespace
-------------------------------------------------------------------------------

local L = LibStub:GetLibrary("SixSocial-1.0", true)
if not L then return end

local Clubs = CreateFrame("Frame")
local Enum = L.Enum
local merge = L.merge

-------------------------------------------------------------------------------
-- Helpers
-------------------------------------------------------------------------------

local function Club_Delete(self)
	local clubId = self[Enum.Club.Id]
	Clubs.clubs[clubId] = nil
end

local function Club_Initialize(self)
	local clubId = self[Enum.Club.Id]
	local isGuild = self[Enum.Club.Type] == Enum.ClubType.Guild

	Clubs.clubs[clubId] = self

	if isGuild then
		Clubs.guildClubId = clubId
	end
end

local function Club_IterateMembers(self, useCachedOnly)
    return L[useCachedOnly and "IterateCachedClubMembers" or "IterateClubMembers"](L, self[Enum.Club.Id])
end

local function Club_Update(self)
	local clubId = self[Enum.Club.Id]
	local clubInfo = C_Club.GetClubInfo(clubId)
	if (not clubInfo) then return end

	self[Enum.Club.Broadcast] = clubInfo.broadcast
	self[Enum.Club.MemberCount] = clubInfo.memberCount
	self[Enum.Club.Name] = clubInfo.name
end

-------------------------------------------------------------------------------
-- Clubs
-------------------------------------------------------------------------------

function Clubs:Initialize()
	if self.initialized then return end
	self.initialized = true
	self.clubs = {}
	self:SetScript("OnEvent", self.OnEvent)
	self:RegisterEvent("CLUB_ADDED")
	self:RegisterEvent("CLUB_MEMBER_ADDED")
	self:RegisterEvent("CLUB_MEMBER_PRESENCE_UPDATED")
	self:RegisterEvent("CLUB_MEMBER_REMOVED")
	self:RegisterEvent("CLUB_MEMBER_ROLE_UPDATED")
	self:RegisterEvent("CLUB_MEMBER_UPDATED")
	self:RegisterEvent("CLUB_REMOVED")
	self:RegisterEvent("CLUB_UPDATED")
	self:RegisterEvent("GUILD_ROSTER_UPDATE")
	self:RegisterEvent("INITIAL_CLUBS_LOADED")
end

function Clubs:OnEvent(event, clubId)
	local club = self.clubs[(event == "GUILD_ROSTER_UPDATE") and self.guildClubId or clubId]

	if event == "CLUB_ADDED" then
		self:CreateClub(clubId)
	elseif event == "CLUB_REMOVED" and club then
		club:Delete()
	elseif (event == "CLUB_MEMBER_ADDED" or event == "CLUB_MEMBER_REMOVED" or event == "CLUB_UPDATED") and club then
		club:Update()
	elseif (event == "GUILD_ROSTER_UPDATE") and club then
		for member in club:IterateMembers(true) do
			member:Delete()
		end
	end

	if event == "CLUB_ADDED" or event == "CLUB_REMOVED" or event == "INITIAL_CLUBS_LOADED" then
		L:SendMessageWithDelay("SIXSOCIAL_CLUB_UPDATE")
	elseif event == "GUILD_ROSTER_UPDATE" then
		L:SendMessageWithDelay("SIXSOCIAL_CLUB_UPDATE", self.guildClubId)
	else
		L:SendMessageWithDelay("SIXSOCIAL_CLUB_UPDATE", clubId)
	end
end

function Clubs:CreateClub(clubId)
	local clubInfo = C_Club.GetClubInfo(clubId)
	if (not clubInfo) then return end

	local club = merge(L.CLUB, {
		[Enum.Club.Broadcast] = clubInfo.broadcast,
		[Enum.Club.Delete] = Club_Delete,
		[Enum.Club.Id] = clubId,
		[Enum.Club.Initialize] = Club_Initialize,
		[Enum.Club.IterateMembers] = Club_IterateMembers,
		[Enum.Club.MemberCount] = clubInfo.memberCount,
		[Enum.Club.Name] = clubInfo.name,
		[Enum.Club.Type] = clubInfo.clubType,
		[Enum.Club.Update] = Club_Update,
	})

	club:Initialize()

	return club
end

-------------------------------------------------------------------------------
-- Library
-------------------------------------------------------------------------------

function L:GetClub(clubId)
	Clubs:Initialize()
	return Clubs.clubs[clubId] or Clubs:CreateClub(clubId)
end

function L:GetGuildClubId()
	Clubs:Initialize()
	return Clubs.guildClubId
end

function L:IterateClubs()
	local clubs = C_Club.GetSubscribedClubs()

	return coroutine.wrap(function()
		for _, clubData in ipairs(clubs) do
			local club = L:GetClub(clubData.clubId)

			if club then
				coroutine.yield(club)
			end
		end
	end)
end
