
-------------------------------------------------------------------------------
-- Optimization
-------------------------------------------------------------------------------

-- GLOBALS: BNGetInfo, BNet_GetClientEmbeddedAtlas, GetRealZoneText, GetUnitName, select, string, UnitClass, UnitFactionGroup, UnitLevel

local BNET_CLIENT_WOW = BNET_CLIENT_WOW
local BNet_GetClientEmbeddedTexture = BNet_GetClientEmbeddedTexture
local BNGetNumFriends = BNGetNumFriends
local C_BattleNet = C_BattleNet
local C_FriendList = C_FriendList
local C_Texture = C_Texture
local coroutine = coroutine
local FRIENDS = FRIENDS
local table = table
local tonumber = tonumber
local type = type
local UNKNOWN = UNKNOWN

-------------------------------------------------------------------------------
-- Namespace
-------------------------------------------------------------------------------

local L = LibStub:GetLibrary("SixSocial-1.0", true)
if not L then return end

local Enum = L.Enum
local falsy = L.falsy
local FormatUnit = L.FormatUnit
local Friends = CreateFrame("Frame")
local GetFriendCategory = L.GetFriendCategory
local GetFriendNote = L.GetFriendNote
local merge = L.merge

local ACCOUNT_ID_OFFSET = 200
local PLAYER_INDEX = 1

-------------------------------------------------------------------------------
-- Helpers
-------------------------------------------------------------------------------

local function Friend_Initialize(self)
	local accountID = self[Enum.Friend.AccountId]
	local battleTag = self[Enum.Friend.BattleTag]
	local friendIndex = self[Enum.Friend.Index]
	local friends = Friends.FRIEND_TYPE_TO_TABLE[self[Enum.Friend.Type]]
	local unit = self[Enum.Friend.Unit]

	if accountID then friends[ACCOUNT_ID_OFFSET + accountID] = self end
	if battleTag then friends[battleTag] = self end
	if friendIndex then friends[friendIndex] = self end
	if unit then friends[unit] = self end
end

local function Friend_Update(self)
	local friendIndex = self[Enum.Friend.Index]
	local friendType = self[Enum.Friend.Type]

	if friendType == Enum.FriendType.BattleNet then
		Friends:CreateBattleNetFriend(friendIndex)
	elseif friendType == Enum.FriendType.Character then
		Friends:CreateCharacterFriend(friendIndex)
	elseif friendType == Enum.FriendType.Player then
		Friends:CreatePlayer()
	end
end

local function GetClientEmbeddedTexture(clientProgram)
	local result = ""

	if C_Texture.IsTitleIconTextureReady(clientProgram, 0) then
		C_Texture.GetTitleIconTexture(clientProgram, 0, function (success, texture)
			if success and texture then
				result = BNet_GetClientEmbeddedTexture(texture, 0, 0, 0)
			end
		end)
	end

	return result
end

local function IsActive(game)
	return game.clientProgram ~= "App" and game.clientProgram ~= "BSAp"
end

local function GetBattleNetFriendActivity(game, status)
	return (game.clientProgram == BNET_CLIENT_WOW and game.areaName)
		or (IsActive(game) and (falsy(game.richPresence) or Enum.ClientName[game.clientProgram]))
		or Enum.StatusText[status]
		or UNKNOWN
end

local function GetBattleNetFriendStatus(bnet, game)
	if (not game.isOnline) then
		return Enum.Status.Offline
	elseif (game.clientProgram == "BSAp") then
		return Enum.Status.Mobile
	elseif bnet.isAFK then
		return Enum.Status.Away
	elseif bnet.isDND then
		return Enum.Status.Busy
	elseif (not IsActive(game) and game.isGameAFK) then
		return Enum.Status.Away
	elseif (not IsActive(game) and game.isGameBusy) then
		return Enum.Status.Busy
	else
		return Enum.Status.Online
	end
end

local function GetCharacterFriendStatus(friendInfo)
	if not friendInfo.connected then
		return Enum.Status.Offline
	elseif friendInfo.afk then
		return Enum.Status.Away
	elseif friendInfo.dnd then
		return Enum.Status.Busy
	else
		return Enum.Status.Online
	end
end

local function GetPlayerActivity()
	return GetRealZoneText() or UNKNOWN
end

local function GetPlayerStatus(isAFK, isDND)
	if isAFK then
		return Enum.Status.Away
	elseif isDND then
		return Enum.Status.Busy
	else
		return Enum.Status.Online
	end
end

-------------------------------------------------------------------------------
-- Friends
-------------------------------------------------------------------------------

function Friends:Initialize()
	if self.initialized then return end
	self.initialized = true
	self.battleNetFriends = {}
	self.characterFriends = {}
	self.player = {}
	self.FRIEND_TYPE_TO_TABLE = {
		[Enum.FriendType.BattleNet] = self.battleNetFriends,
		[Enum.FriendType.Character] = self.characterFriends,
		[Enum.FriendType.Player] = self.player,
	}
	self:SetScript("OnEvent", self.OnEvent)
	self:RegisterEvent("BN_CONNECTED")
	self:RegisterEvent("BN_CUSTOM_MESSAGE_CHANGED")
	self:RegisterEvent("BN_CUSTOM_MESSAGE_LOADED")
	self:RegisterEvent("BN_DISCONNECTED")
	self:RegisterEvent("BN_FRIEND_INFO_CHANGED")
	self:RegisterEvent("BN_FRIEND_INVITE_LIST_INITIALIZED")
	self:RegisterEvent("BN_FRIEND_LIST_SIZE_CHANGED")
	self:RegisterEvent("BN_INFO_CHANGED")
	self:RegisterEvent("FRIENDLIST_UPDATE")
	self:RegisterEvent("PLAYER_ENTERING_WORLD")
	self:CreateFriends(Enum.FriendType.Any)
end

function Friends:OnEvent(event, friendIndex)
	local friend = friendIndex and self.battleNetFriends[friendIndex]
	local player = self.player[PLAYER_INDEX]

	if (event == "BN_CUSTOM_MESSAGE_CHANGED" or event == "BN_FRIEND_INFO_CHANGED") and friend then
		friend:Update()
	elseif (event == "BN_CUSTOM_MESSAGE_CHANGED" or event == "BN_CUSTOM_MESSAGE_LOADED" or event == "BN_INFO_CHANGED") and player then
		player:Update()
	elseif event == "FRIENDLIST_UPDATE" then
		self:CreateFriends(Enum.FriendType.Character)
	elseif event == "PLAYER_ENTERING_WORLD" then
		self:CreateFriends(Enum.FriendType.Any)
	else
		self:CreateFriends(Enum.FriendType.BattleNet)
	end

	L:SendMessageWithDelay("SIXSOCIAL_BNFRIEND_UPDATE")
end

function Friends:CreateFriends(friendType)
	if friendType == Enum.FriendType.Any or friendType == Enum.FriendType.BattleNet then
		table.wipe(self.battleNetFriends)

		for friendIndex = 1, (BNGetNumFriends() or 0) do
			self:CreateBattleNetFriend(friendIndex)
		end
	end

	if friendType == Enum.FriendType.Any or friendType == Enum.FriendType.Character then
		table.wipe(self.characterFriends)

		for friendIndex = 1, (C_FriendList.GetNumFriends() or 0) do
			self:CreateCharacterFriend(friendIndex)
		end
	end

	if friendType == Enum.FriendType.Any or friendType == Enum.FriendType.Player then
		table.wipe(self.player)
		self:CreatePlayer()
	end
end

function Friends:CreateBattleNetFriend(friendIndex)
	local bnet = C_BattleNet.GetFriendAccountInfo(friendIndex)
	if not bnet then return end
	local game = bnet.gameAccountInfo
	local status = GetBattleNetFriendStatus(bnet, game)
	local unit = game.clientProgram == BNET_CLIENT_WOW and FormatUnit(game.characterName, game.realmName, game.richPresence)

	local friend = merge(L.FRIEND, {
		[Enum.Friend.AccountId] = bnet.bnetAccountID,
		[Enum.Friend.AccountName] = bnet.accountName,
		[Enum.Friend.Activity] = GetBattleNetFriendActivity(game, status),
		[Enum.Friend.BattleTag] = bnet.battleTag,
		[Enum.Friend.Broadcast] = falsy(bnet.customMessage),
		[Enum.Friend.Category] = GetFriendCategory(Enum.FriendType.BattleNet, bnet.note, bnet.isFavorite),
		[Enum.Friend.ClientName] = game.clientProgram,
		[Enum.Friend.ClientTexture] = GetClientEmbeddedTexture(game.clientProgram),
		[Enum.Friend.Index] = friendIndex,
		[Enum.Friend.Initialize] = Friend_Initialize,
		[Enum.Friend.IsActive] = IsActive(game),
		[Enum.Friend.IsFavorite] = bnet.isFavorite,
		[Enum.Friend.Name] = bnet.accountName,
		[Enum.Friend.Note] = GetFriendNote(bnet.note),
		[Enum.Friend.Status] = status,
		[Enum.Friend.StatusOrder] = Enum.StatusOrder[status],
		[Enum.Friend.Type] = Enum.FriendType.BattleNet,
		[Enum.Friend.Update] = Friend_Update,
	}, unit and {
		[Enum.Friend.CharacterName] = falsy(game.characterName),
		[Enum.Friend.ClassName] = game.className,
		[Enum.Friend.FactionName] = game.factionName,
		[Enum.Friend.Level] = falsy(tonumber(game.characterLevel)),
		[Enum.Friend.Unit] = unit,
	} or {})

	friend:Initialize()

	return friend
end

function Friends:CreateCharacterFriend(friendIndex)
	local info = C_FriendList.GetFriendInfoByIndex(friendIndex)
	if (not info) then return end
	local status = GetCharacterFriendStatus(info)
	local unit = FormatUnit(info.name)

	local friend = merge(L.FRIEND, {
		[Enum.Friend.Activity] = Enum.StatusText[status],
		[Enum.Friend.Category] = GetFriendCategory(Enum.FriendType.Character, info.notes, false),
		[Enum.Friend.CharacterName] = falsy(info.name),
		[Enum.Friend.ClientName] = Enum.ClientName[BNET_CLIENT_WOW],
		[Enum.Friend.ClientTexture] = GetClientEmbeddedTexture(BNET_CLIENT_WOW),
		[Enum.Friend.FriendGuid] = info.guid,
		[Enum.Friend.Index] = friendIndex,
		[Enum.Friend.Initialize] = Friend_Initialize,
		[Enum.Friend.IsActive] = info.connected,
		[Enum.Friend.Note] = GetFriendNote(info.notes),
		[Enum.Friend.Status] = status,
		[Enum.Friend.StatusOrder] = Enum.StatusOrder[status],
		[Enum.Friend.Type] = Enum.FriendType.Character,
		[Enum.Friend.Update] = Friend_Update,
	}, status ~= Enum.Status.Offline and {
		[Enum.Friend.Activity] = info.area,
		[Enum.Friend.ClassName] = info.className,
		[Enum.Friend.FactionName] = L.PLAYER_FACTION_GROUP,
		[Enum.Friend.Level] = falsy(info.level),
		[Enum.Friend.Unit] = unit,
	} or {})

	friend:Initialize()

	return friend
end

function Friends:CreatePlayer()
	local _, battleTag, accountID, customMessage, isAFK, isDND = BNGetInfo()
	if (not accountID) then return end
	local name = string.split("#", battleTag)
	local status = GetPlayerStatus(isAFK, isDND)
	local unit = GetUnitName("player")

	local player = merge(L.FRIEND, {
		[Enum.Friend.AccountId] = accountID,
		[Enum.Friend.AccountName] = name,
		[Enum.Friend.Activity] = GetPlayerActivity(),
		[Enum.Friend.BattleTag] = battleTag,
		[Enum.Friend.Broadcast] = falsy(customMessage),
		[Enum.Friend.Category] = FRIENDS,
		[Enum.Friend.CharacterName] = unit,
		[Enum.Friend.ClassName] = select(2, UnitClass("player")),
		[Enum.Friend.ClientName] = BNET_CLIENT_WOW,
		[Enum.Friend.ClientTexture] = BNet_GetClientEmbeddedAtlas(BNET_CLIENT_WOW, 0),
		[Enum.Friend.FactionName] = UnitFactionGroup("player"),
		[Enum.Friend.Index] = PLAYER_INDEX,
		[Enum.Friend.Initialize] = Friend_Initialize,
		[Enum.Friend.IsActive] = true,
		[Enum.Friend.Level] = UnitLevel("player"),
		[Enum.Friend.Name] = name,
		[Enum.Friend.Status] = status,
		[Enum.Friend.StatusOrder] = Enum.StatusOrder[status],
		[Enum.Friend.Type] = Enum.FriendType.Player,
		[Enum.Friend.Unit] = unit,
		[Enum.Friend.Update] = Friend_Update,
	})

	player:Initialize()

	return player
end

-------------------------------------------------------------------------------
-- Library
-------------------------------------------------------------------------------

function L:GetFriend(friendType, friendIndex)
	Friends:Initialize()

	if friendType == Enum.FriendType.BattleNet then
		return Friends.battleNetFriends[friendIndex]
	elseif friendType == Enum.FriendType.Character then
		return Friends.characterFriends[friendIndex]
	elseif friendType == Enum.FriendType.Player then
		return Friends.player[PLAYER_INDEX]
	end
end

function L:GetFriendByAccountID(accountID)
	if type(accountID) == "number" then
		Friends:Initialize()
		return Friends.battleNetFriends[ACCOUNT_ID_OFFSET + accountID] or Friends.player[ACCOUNT_ID_OFFSET + accountID]
	end
end

function L:GetFriendByBattleTag(battleTag)
	Friends:Initialize()
	return Friends.battleNetFriends[battleTag] or Friends.player[battleTag]
end

function L:GetFriendByUnit(unit)
	Friends:Initialize()
	return Friends.battleNetFriends[unit] or Friends.characterFriends[unit] or Friends.player[unit]
end

function L:IterateFriends()
	local numBattleNetFriends = BNGetNumFriends() or 0
	local numCharacterFriends = C_FriendList.GetNumFriends() or 0
	local numFriends = numBattleNetFriends + numCharacterFriends

	return coroutine.wrap(function()
		for friendIndex = 1, numFriends do
			local friend = nil

			if friendIndex <= numBattleNetFriends then
				friend = self:GetFriend(Enum.FriendType.BattleNet, friendIndex)
			else
				friend = self:GetFriend(Enum.FriendType.Character, friendIndex - numBattleNetFriends)
			end

			if friend then
				coroutine.yield(friend)
			end
		end
	end)
end

function L:WrapUnitInFriend(unit)
	return merge(L.FRIEND, {unit = unit})
end
